"use strict";
/*
 *  This file is part of CoCalc: Copyright © 2020 Sagemath, Inc.
 *  License: AGPLv3 s.t. "Commons Clause" – see LICENSE.md for details
 */
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.from_str = exports.DBDocument = void 0;
var immutable = require("immutable");
var underscore_1 = require("underscore");
var misc_1 = require("../../../misc");
var util_1 = require("../generic/util");
var util_2 = require("./util");
// Immutable DB document
var DBDocument = /** @class */ (function () {
    function DBDocument(primary_keys, string_cols, records, everything, indexes, change_tracker) {
        if (records === void 0) { records = immutable.List(); }
        this.set = this.set.bind(this);
        this.delete_array = this.delete_array.bind(this);
        this.primary_key_cols = this.primary_key_cols.bind(this);
        this.primary_key_part = this.primary_key_part.bind(this);
        this.primary_keys = new Set(primary_keys);
        if (this.primary_keys.size === 0) {
            throw Error("there must be at least one primary key");
        }
        this.string_cols = new Set(string_cols);
        this.records = records;
        this.everything = everything == null ? this.init_everything() : everything;
        this.size = this.everything.size;
        this.indexes = indexes == null ? this.init_indexes() : indexes;
        this.change_tracker =
            change_tracker == null ? this.init_change_tracker() : change_tracker;
    }
    // sorted immutable Set of i such that this.records.get(i) is defined.
    DBDocument.prototype.init_everything = function () {
        var v = [];
        for (var n = 0; n < this.records.size; n++) {
            if (this.records.get(n) != undefined) {
                v.push(n);
            }
        }
        return immutable.Set(v);
    };
    DBDocument.prototype.init_indexes = function () {
        var e_1, _a;
        // Build indexes
        var indexes = immutable.Map();
        try {
            for (var _b = __values(this.primary_keys), _c = _b.next(); !_c.done; _c = _b.next()) {
                var field = _c.value;
                var index = immutable.Map();
                indexes = indexes.set(field, index);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
        this.records.map(function (record, n) {
            if (record == null) {
                // null records are sentinels for deletions.
                return;
            }
            indexes.map(function (index, field) {
                var val = record.get(field);
                if (val != null) {
                    var k = util_2.to_key(val);
                    var matches = index.get(k);
                    if (matches != null) {
                        matches = matches.add(n);
                    }
                    else {
                        matches = immutable.Set([n]);
                    }
                    indexes = indexes.set(field, index.set(k, matches));
                }
            });
        });
        return indexes;
    };
    DBDocument.prototype.init_change_tracker = function () {
        return { changes: immutable.Set(), from_db: this };
    };
    DBDocument.prototype.to_str = function () {
        if (this.to_str_cache != null) {
            // We can use a cache, since this is an immutable object
            return this.to_str_cache;
        }
        var obj = this.get({}).toJS();
        return (this.to_str_cache = util_2.to_str(obj));
    };
    DBDocument.prototype.is_equal = function (other) {
        if (other == null) {
            // Definitely not equal if not defined.
            return false;
        }
        if (this.records === other.records) {
            return true;
        }
        if (this.size !== other.size) {
            return false;
        }
        // We include undefineds in the sets below
        // since records is a List of Record or undefined, i.e.,
        // we use undefined as a sentinel in order to
        // make deletes be efficient.
        return immutable
            .Set(this.records)
            .add(undefined)
            .equals(immutable.Set(other.records).add(undefined));
    };
    DBDocument.prototype.apply_patch = function (patch) {
        var i = 0;
        var db = this;
        while (i < patch.length) {
            if (patch[i] === -1) {
                db = db.delete(patch[i + 1]);
            }
            else if (patch[i] === 1) {
                db = db.set(patch[i + 1]);
            }
            i += 2;
        }
        return db;
    };
    DBDocument.prototype.make_patch = function (other) {
        var _this = this;
        if (other.size === 0) {
            // Special case -- delete everything
            return [-1, [{}]];
        }
        var t0 = immutable.Set(this.records);
        var t1 = immutable.Set(other.records);
        // Remove the common intersection -- nothing going on there.
        // Doing this greatly reduces the complexity in the common
        // case in which little has changed
        var common = t0.intersect(t1).add(undefined);
        t0 = t0.subtract(common);
        t1 = t1.subtract(common);
        // Easy very common special cases
        if (t0.size === 0) {
            // Special case: t0 is empty -- insert all the records.
            return [1, t1.toJS()];
        }
        if (t1.size === 0) {
            // Special case: t1 is empty -- bunch of deletes
            var v_1 = [];
            t0.map(function (x) {
                if (x != null) {
                    v_1.push(_this.primary_key_part(x.toJS()));
                }
            });
            return [-1, v_1];
        }
        // compute the key parts of t0 and t1 as sets
        // means -- set got from t0 by taking only the primary_key columns
        // (Why the "as"? Typescript makes the output of the map be of type
        //        Iterable<Record, Iterable<string, any>>
        // But, it's just a set.  So cast it.)
        var k0 = t0.map(this.primary_key_cols);
        var k1 = t1.map(this.primary_key_cols);
        var add = [];
        var remove = undefined;
        // Deletes: everything in k0 that is not in k1
        var deletes = k0.subtract(k1);
        if (deletes.size > 0) {
            remove = deletes.toJS();
        }
        // Inserts: everything in k1 that is not in k0
        var inserts = k1.subtract(k0);
        if (inserts.size > 0) {
            inserts.map(function (k) {
                if (k != null) {
                    var x = other.get_one(k);
                    if (x != null) {
                        add.push(x.toJS());
                    }
                }
            });
        }
        // Everything in k1 that is also in k0 -- these
        // must have all changed
        var changed = k1.intersect(k0);
        if (changed.size > 0) {
            changed.map(function (k) {
                if (k == null) {
                    return;
                }
                var obj = k.toJS();
                var obj0 = _this.primary_key_part(obj);
                var from0 = _this.get_one(obj0);
                var to0 = other.get_one(obj0);
                if (from0 == null || to0 == null) {
                    // just to make typescript happy
                    return;
                }
                var from = from0.toJS();
                var to = to0.toJS();
                // undefined for each key of from not in to
                for (var key in from) {
                    if (to[key] == null) {
                        obj[key] = null;
                    }
                }
                // Explicitly set each key of `to` that is different
                // than corresponding key of `from`:
                for (var key in to) {
                    var v = to[key];
                    if (!underscore_1.isEqual(from[key], v)) {
                        if (_this.string_cols.has(key) && from[key] != null && v != null) {
                            if (typeof from[key] == "string" && typeof v == "string") {
                                // A valid string patch, converting one string to another.
                                obj[key] = util_1.make_patch(from[key], v);
                            }
                            else {
                                /* This should be impossible, due to the type checking that
                                   I've added to set in this same commit.  However, it can't
                                   hurt to put in the above check on types, just in case.
                                     https://github.com/sagemathinc/cocalc/issues/3625
                                   A string col actually contains something that is not
                                   a string.  (Maybe it's due to loading something old?)
                                   In any case, it's better to "best effort" this, rather
                                   than to make the entire document be un-savable and
                                   un-usable to the user.
                                   We just give up and record no change in this case, so
                                   when doc is read in later (or by another user), there
                                   will be no weird corruption.
                                   This case will probably go away completely when all
                                   client code is written with proper typing.
                                */
                            }
                        }
                        else if (misc_1.is_object(from[key]) && misc_1.is_object(v)) {
                            // Changing from one map to another, where they are not
                            // equal -- can use a merge to make this more efficient.
                            // This is an important optimization, to avoid making
                            // patches HUGE.
                            obj[key] = util_2.map_merge_patch(from[key], v);
                        }
                        else {
                            obj[key] = v;
                        }
                    }
                }
                add.push(obj);
            });
        }
        var patch = [];
        if (remove != null) {
            patch.push(-1);
            patch.push(remove);
        }
        if (add.length > 0) {
            patch.push(1);
            patch.push(add);
        }
        return patch;
    };
    // Given an immutable map f, return its restriction
    // to the primary keys.
    DBDocument.prototype.primary_key_cols = function (f) {
        var _this = this;
        return f.filter(function (_, k) { return k != null && _this.primary_keys.has(k); });
    };
    DBDocument.prototype.select = function (where) {
        if (immutable.Map.isMap(where)) {
            // TODO: maybe do not allow?
            where = where.toJS();
        }
        // Return immutable set with defined indexes the elts of @_records that
        // satisfy the where condition.
        var n = misc_1.len(where);
        var result = undefined;
        for (var field in where) {
            var value = where[field];
            var index = this.indexes.get(field);
            if (index == null) {
                throw Error("field '" + field + "' must be a primary key");
            }
            var v = index.get(util_2.to_key(value));
            // v may be undefined here
            if (v == null) {
                return immutable.Set(); // no matches for this field - done
            }
            if (n === 1) {
                // no need to do further intersection
                return v;
            }
            if (result != null) {
                // intersect with what we've found so far via indexes.
                result = result.intersect(v);
            }
            else {
                result = v;
            }
        }
        if (result == null) {
            // where condition must have been empty -- matches everything
            return this.everything;
        }
        else {
            return result;
        }
    };
    // Used internally for determining the set/where parts of an object.
    DBDocument.prototype.parse = function (obj) {
        var where = {};
        var set = {};
        for (var field in obj) {
            var val = obj[field];
            if (this.primary_keys.has(field)) {
                if (val != null) {
                    where[field] = val;
                }
            }
            else {
                set[field] = val;
            }
        }
        return { where: where, set: set };
    };
    DBDocument.prototype.set = function (obj) {
        var e_2, _a, e_3, _b, e_4, _c;
        if (misc_1.is_array(obj)) {
            var z = this;
            try {
                for (var _d = __values(obj), _e = _d.next(); !_e.done; _e = _d.next()) {
                    var x = _e.value;
                    z = z.set(x);
                }
            }
            catch (e_2_1) { e_2 = { error: e_2_1 }; }
            finally {
                try {
                    if (_e && !_e.done && (_a = _d.return)) _a.call(_d);
                }
                finally { if (e_2) throw e_2.error; }
            }
            return z;
        }
        if (immutable.Map.isMap(obj)) {
            // TODO: maybe do not allow?
            // it is very clean/convenient to allow this
            obj = obj.toJS();
        }
        var _f = this.parse(obj), set = _f.set, where = _f.where;
        var matches = this.select(where);
        var changes = this.change_tracker.changes;
        var first_match = matches != null ? matches.min() : undefined;
        if (first_match != null) {
            // edit the first existing record that matches
            var record = this.records.get(first_match);
            if (record == null) {
                // make typescript happier.
                throw Error("bug -- record can't be null");
            }
            var before = record;
            for (var field in set) {
                var value = set[field];
                if (value === null) {
                    // null = how to delete fields
                    record = record.delete(field);
                }
                else {
                    if (this.string_cols.has(field)) {
                        if (misc_1.is_array(value)) {
                            // special case: a string patch
                            record = record.set(field, util_1.apply_patch(value, before.get(field, ""))[0]);
                            continue;
                        }
                        if (typeof value != "string") {
                            // Putting this guard in to address
                            // https://github.com/sagemathinc/cocalc/issues/3625
                            // which was caused by some client code setting a string_col
                            // to something that is not a string.  We'll next have
                            // to wait for this exception to be triggered someday...
                            throw Error("'" + field + "' must be a string, but tried to set to '" + value + "' of type " + typeof value);
                        }
                        // falls through to actually set it below.
                    }
                    var new_val = void 0;
                    var cur = record.get(field);
                    var change = immutable.fromJS(value);
                    if (immutable.Map.isMap(cur) && immutable.Map.isMap(change)) {
                        new_val = util_2.merge_set(cur, change);
                    }
                    else {
                        new_val = change;
                    }
                    record = record.set(field, new_val);
                }
            }
            if (!before.equals(record)) {
                // there was an actual change, so update; doesn't change
                // anything involving indexes.
                changes = changes.add(this.primary_key_cols(record));
                return new DBDocument(this.primary_keys, this.string_cols, this.records.set(first_match, record), this.everything, this.indexes, { changes: changes, from_db: this.change_tracker.from_db });
            }
            else {
                return this;
            }
        }
        else {
            try {
                // The sparse array matches had nothing in it, so
                // append a new record.
                for (var _g = __values(this.string_cols), _h = _g.next(); !_h.done; _h = _g.next()) {
                    var field = _h.value;
                    if (obj[field] != null && misc_1.is_array(obj[field])) {
                        // It's a patch -- but there is nothing to patch,
                        // so discard this field
                        obj = misc_1.copy_without(obj, field);
                    }
                }
            }
            catch (e_3_1) { e_3 = { error: e_3_1 }; }
            finally {
                try {
                    if (_h && !_h.done && (_b = _g.return)) _b.call(_g);
                }
                finally { if (e_3) throw e_3.error; }
            }
            // remove null columns (null indicates delete)
            var record = util_2.nonnull_cols(immutable.fromJS(obj));
            changes = changes.add(this.primary_key_cols(record));
            var records = this.records.push(record);
            var n = records.size - 1;
            var everything = this.everything.add(n);
            // update indexes
            var indexes = this.indexes;
            try {
                for (var _j = __values(this.primary_keys), _k = _j.next(); !_k.done; _k = _j.next()) {
                    var field = _k.value;
                    var val = obj[field];
                    if (val != null) {
                        var index = indexes.get(field);
                        if (index == null) {
                            index = immutable.Map();
                        }
                        var k = util_2.to_key(val);
                        var matches_1 = index.get(k);
                        if (matches_1 != null) {
                            matches_1 = matches_1.add(n);
                        }
                        else {
                            matches_1 = immutable.Set([n]);
                        }
                        indexes = indexes.set(field, index.set(k, matches_1));
                    }
                }
            }
            catch (e_4_1) { e_4 = { error: e_4_1 }; }
            finally {
                try {
                    if (_k && !_k.done && (_c = _j.return)) _c.call(_j);
                }
                finally { if (e_4) throw e_4.error; }
            }
            return new DBDocument(this.primary_keys, this.string_cols, records, everything, indexes, { changes: changes, from_db: this.change_tracker.from_db });
        }
    };
    DBDocument.prototype.delete_array = function (where) {
        var e_5, _a;
        var z = this;
        try {
            for (var where_1 = __values(where), where_1_1 = where_1.next(); !where_1_1.done; where_1_1 = where_1.next()) {
                var x = where_1_1.value;
                z = z.delete(x);
            }
        }
        catch (e_5_1) { e_5 = { error: e_5_1 }; }
        finally {
            try {
                if (where_1_1 && !where_1_1.done && (_a = where_1.return)) _a.call(where_1);
            }
            finally { if (e_5) throw e_5.error; }
        }
        return z;
    };
    DBDocument.prototype.delete = function (where) {
        var e_6, _a;
        var _this = this;
        // console.log("delete #{JSON.stringify(where)}")
        if (misc_1.is_array(where)) {
            return this.delete_array(where);
        }
        // if where undefined, will delete everything
        if (this.everything.size === 0) {
            // no-op -- no data so deleting is trivial
            return this;
        }
        var changes = this.change_tracker.changes;
        var remove = this.select(where);
        if (remove.size === this.everything.size) {
            // actually deleting everything; easy special cases
            changes = changes.union(this.records
                .filter(function (record) { return record != null; })
                .map(this.primary_key_cols));
            return new DBDocument(this.primary_keys, this.string_cols, undefined, undefined, undefined, { changes: changes, from_db: this.change_tracker.from_db });
        }
        // remove matches from every index
        var indexes = this.indexes;
        var _loop_1 = function (field) {
            var index = indexes.get(field);
            if (index == null) {
                return "continue";
            }
            remove.forEach(function (n) {
                if (n == null) {
                    return;
                }
                var record = _this.records.get(n);
                if (record == null) {
                    return;
                }
                var val = record.get(field);
                if (val == null) {
                    return;
                }
                var k = util_2.to_key(val);
                var matches = index.get(k).delete(n);
                if (matches.size === 0) {
                    index = index.delete(k);
                }
                else {
                    index = index.set(k, matches);
                }
                indexes = indexes.set(field, index);
            });
        };
        try {
            for (var _b = __values(this.primary_keys), _c = _b.next(); !_c.done; _c = _b.next()) {
                var field = _c.value;
                _loop_1(field);
            }
        }
        catch (e_6_1) { e_6 = { error: e_6_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_6) throw e_6.error; }
        }
        // delete corresponding records (actually set to undefined to
        // preserve index references).
        var records = this.records;
        remove.forEach(function (n) {
            if (n == null) {
                return;
            }
            var record = records.get(n);
            if (record == null) {
                return;
            }
            changes = changes.add(_this.primary_key_cols(record));
            records = records.set(n, undefined);
        });
        var everything = this.everything.subtract(remove);
        return new DBDocument(this.primary_keys, this.string_cols, records, everything, indexes, { changes: changes, from_db: this.change_tracker.from_db });
    };
    // Returns immutable list of all matches
    DBDocument.prototype.get = function (where) {
        var matches = this.select(where);
        if (matches == null) {
            return immutable.List();
        }
        // The "as" is because Typescript just makes the result of
        // filter some more generic type (but it isn't).
        return immutable.List(this.records.filter(function (_, n) { return n != null && matches.includes(n); }));
    };
    // Returns the first match, or undefined if there are no matches
    DBDocument.prototype.get_one = function (where) {
        // TODO: couldn't select have a shortcut to return once one
        // result is found.
        var matches = this.select(where);
        if (matches == null) {
            return;
        }
        return this.records.get(matches.min());
    };
    // x = javascript object
    DBDocument.prototype.primary_key_part = function (x) {
        var where = {};
        for (var k in x) {
            var v = x[k];
            if (this.primary_keys.has(k)) {
                where[k] = v;
            }
        }
        return where;
    };
    // Return immutable set of primary key parts of records that
    // change in going from this to other.
    DBDocument.prototype.changed_keys = function (other) {
        if (this.records === other.records) {
            // identical -- obviously, nothing changed.
            return immutable.Set();
        }
        // Get the defined records; there may be undefined ones due
        // to lazy delete.
        var t0 = immutable.Set(immutable.Set(this.records).filter(function (x) { return x != null; }));
        var t1 = immutable.Set(immutable.Set(other.records).filter(function (x) { return x != null; }));
        // Remove the common intersection -- nothing going on there.
        // Doing this greatly reduces the complexity in the common
        // case in which little has changed
        var common = t0.intersect(t1);
        t0 = t0.subtract(common);
        t1 = t1.subtract(common);
        // compute the key parts of t0 and t1 as sets
        var k0 = immutable.Set(t0.map(this.primary_key_cols));
        var k1 = immutable.Set(t1.map(this.primary_key_cols));
        return immutable.Set(k0.union(k1));
    };
    DBDocument.prototype.changes = function (prev) {
        // CRITICAL TODO!  Make this efficient using this.change_tracker!!!
        if (prev == null) {
            return immutable.Set(immutable
                .Set(this.records)
                .filter(function (x) { return x != null; })
                .map(this.primary_key_cols));
        }
        return this.changed_keys(prev);
    };
    DBDocument.prototype.count = function () {
        return this.size;
    };
    return DBDocument;
}());
exports.DBDocument = DBDocument;
/*
The underlying string representation has one JSON object
per line.  The order doesn't matter.

WARNING: The primary keys and string cols are NOT stored
in the string representation!  That is metadata that must
somehow be tracked separately.  (Maybe this should be changed).

You can't store null since null is used to signify deleting
(in set queries). You can't store undefined or Date objects
due to JSON.

Also, note that the primary_keys and string_cols are string[]
rather than Set of strings, which is annoyingly inconsistent
with DBDocument above.
*/
function from_str(s, primary_keys, string_cols) {
    var e_7, _a;
    var obj = [];
    try {
        for (var _b = __values(s.split("\n")), _c = _b.next(); !_c.done; _c = _b.next()) {
            var line = _c.value;
            if (line.length > 0) {
                try {
                    var x = JSON.parse(line);
                    if (typeof x === "object") {
                        obj.push(x);
                    }
                    else {
                        throw Error("each line must be an object");
                    }
                }
                catch (e) {
                    console.warn("CORRUPT db-doc string: " + e + " -- skipping '" + line + "'");
                }
            }
        }
    }
    catch (e_7_1) { e_7 = { error: e_7_1 }; }
    finally {
        try {
            if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
        }
        finally { if (e_7) throw e_7.error; }
    }
    return new DBDocument(new Set(primary_keys), new Set(string_cols), immutable.fromJS(obj));
}
exports.from_str = from_str;
