"use strict";
/*
 *  This file is part of CoCalc: Copyright © 2020 Sagemath, Inc.
 *  License: AGPLv3 s.t. "Commons Clause" – see LICENSE.md for details
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.nonnull_cols = exports.merge_set = exports.map_merge_patch = exports.to_str = exports.to_key = void 0;
// Well-defined JSON.stringify...
var json_stable = require("json-stable-stringify");
var immutable = require("immutable");
var underscore_1 = require("underscore");
function to_key(s) {
    if (immutable.Map.isMap(s)) {
        s = s.toJS();
    }
    // NOTE: s is not undefined.
    return json_stable(s);
}
exports.to_key = to_key;
/* TODO/worry: I change to json_stable from misc.to_json
   and misc.from_json, so the string is canonical.  However,
   Date objects will be treated differently.   This is fine
   by me, in that probably we should just ensure no date
   objects are ever used with db-doc... or have a special
   column type (like string_cols) for them.
   I think right now all our applications (e.g., task lists, jupyter)
   just use ms since epoch explicitly.
*/
function to_str(obj) {
    var v = obj.map(function (x) { return json_stable(x); });
    /* NOTE: It is *VERY* important to sort v!  Otherwise, the hash
       of this document, which is used by
       syncstring, isn't stable in terms of the value of the
       document.  This can in theory
       cause massive trouble with file saves, e.g., of jupyter
       notebooks, courses, etc. (They save fine, but
       they appear not to for the user...).
     */
    v.sort();
    return v.join("\n");
}
exports.to_str = to_str;
// Create an object change such that merge_set(obj1, change) produces obj2.
// Thus for each key, value1 of obj1 and key, value2 of obj2:
//  If value1 is the same as value2, do nothing.
//  If value1 exists but value2 does not, do change[key] = null
//  If value2 exists but value1 does not, do change[key] = value2
function map_merge_patch(obj1, obj2) {
    var val2;
    var change = {};
    for (var key in obj1) {
        var val1 = obj1[key];
        val2 = obj2[key];
        if (underscore_1.isEqual(val1, val2)) {
            // nothing to do
        }
        else if (val2 == null) {
            change[key] = null;
        }
        else {
            change[key] = val2;
        }
    }
    for (key in obj2) {
        val2 = obj2[key];
        if (obj1[key] != null) {
            continue;
        }
        change[key] = val2;
    }
    return change;
}
exports.map_merge_patch = map_merge_patch;
// obj and change are both immutable.js Maps.  Do the following:
//  - for each value of change that is null or undefined, we delete that key from obj
//  - we set the other vals of obj, accordingly.
// So this is a shallow merge with the ability to *delete* keys.
function merge_set(obj, change) {
    change.forEach(function (v, k) {
        if (v === null || v == null) {
            obj = obj.delete(k);
        }
        else {
            obj = obj.set(k, v);
        }
    });
    return obj;
}
exports.merge_set = merge_set;
function nonnull_cols(f) {
    // Yes, "!==" not "!=" below!
    return immutable.Map(f.filter(function (v, _) { return v !== null; }));
}
exports.nonnull_cols = nonnull_cols;
