"use strict";
/*
 *  This file is part of CoCalc: Copyright © 2020 Sagemath, Inc.
 *  License: AGPLv3 s.t. "Commons Clause" – see LICENSE.md for details
 */
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PatchValueCache = void 0;
/*
The PatchValueCache is used to cache values returned
by SortedPatchList.value.  Caching is critical, since otherwise
the client may have to apply hundreds of patches after ever
few keystrokes, which would make CoCalc unusable.  Also, the
history browser is very painful to use without caching.
*/
var misc_1 = require("../../../misc");
var PatchValueCache = /** @class */ (function () {
    function PatchValueCache() {
        // keys in the cache are integers converted to strings.
        this.cache = {};
    }
    // Remove everything from the value cache that has timestamp >= time.
    // If time not defined, removes everything, thus emptying the cache.
    PatchValueCache.prototype.invalidate = function (time) {
        var time0 = time.valueOf();
        for (var time_1 in this.cache) {
            if (parseInt(time_1) >= time0) {
                delete this.cache[time_1];
            }
        }
    };
    // Ensure the value cache doesn't have too many entries in it by
    // removing all but n of the ones that have not been accessed recently.
    PatchValueCache.prototype.prune = function (n) {
        var e_1, _a;
        if (this.size() <= n) {
            // nothing to do
            return;
        }
        var v = [];
        for (var time in this.cache) {
            var x = this.cache[time];
            if (x != null) {
                v.push({ time: time, last_used: x.last_used });
            }
        }
        v.sort(function (a, b) { return misc_1.cmp_Date(a.last_used, b.last_used); });
        try {
            // Delete oldest n entries.
            for (var _b = __values(v.slice(0, v.length - n)), _c = _b.next(); !_c.done; _c = _b.next()) {
                var x = _c.value;
                delete this.cache[x.time];
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
    };
    // Include the given value at the given point in time, which should be
    // the output of this.value(time), and should involve applying all patches
    // up to this.patches[start-1].
    PatchValueCache.prototype.include = function (time, value, start) {
        this.cache["" + time.valueOf()] = {
            time: time,
            value: value,
            start: start,
            last_used: new Date(),
        };
    };
    PatchValueCache.prototype.keys = function () {
        return misc_1.keys(this.cache).map(function (x) { return parseInt(x); });
    };
    /* Return the newest value x with x.time <= time in the cache as an object
  
          x={time:time, value:value, start:start},
  
       where this.value(time) is the given value, and it was obtained
       by applying the elements of this.patches up to this.patches[start-1]
       Return undefined if there are no cached values.
       If time is undefined, returns the newest value in the cache.
       If strict is true, returns newest value at time strictly older than time
    */
    PatchValueCache.prototype.newest_value_at_most = function (time, strict) {
        var e_2, _a;
        if (strict === void 0) { strict = false; }
        var v = this.keys();
        if (v.length === 0) {
            return;
        }
        v.sort(misc_1.cmp);
        v.reverse();
        if (time == null) {
            return this.get(v[0]);
        }
        var time0 = time.valueOf();
        try {
            for (var v_1 = __values(v), v_1_1 = v_1.next(); !v_1_1.done; v_1_1 = v_1.next()) {
                var t = v_1_1.value;
                if ((!strict && t <= time0) || (strict && t < time0)) {
                    return this.get(t);
                }
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (v_1_1 && !v_1_1.done && (_a = v_1.return)) _a.call(v_1);
            }
            finally { if (e_2) throw e_2.error; }
        }
    };
    /* Return cached entry corresponding to the given point in time.
       Here time must be either a new Date() object, or a number (ms since epoch).
       If there is nothing in the cache for the given time, returns undefined.
       ** YOU BETTER NOT mutate the returned value! **  It's not a copy!!
    */
    PatchValueCache.prototype.get = function (time) {
        if (typeof time !== "number") {
            // also allow dates
            time = time.valueOf();
        }
        var x = this.cache[time];
        if (x == null) {
            return;
        }
        x.last_used = new Date(); // this is only for the client cache, so fine to use browser's clock
        return x;
    };
    PatchValueCache.prototype.oldest_time = function () {
        var v = this.keys();
        if (v.length === 0) {
            return;
        }
        v.sort(misc_1.cmp);
        return new Date(v[0]);
    };
    // Number of cached values
    PatchValueCache.prototype.size = function () {
        return misc_1.len(this.cache);
    };
    return PatchValueCache;
}());
exports.PatchValueCache = PatchValueCache;
