"use strict";
/*
 *  This file is part of CoCalc: Copyright © 2020 Sagemath, Inc.
 *  License: AGPLv3 s.t. "Commons Clause" – see LICENSE.md for details
 */
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SortedPatchList = void 0;
/*
Sorted list of patches applied to a starting string/object.
*/
var events_1 = require("events");
var underscore_1 = require("underscore");
var util_1 = require("./util");
var misc_1 = require("../../../misc");
var patch_value_cache_1 = require("./patch-value-cache");
// Make this bigger to make things faster... at the
// cost of using more memory.
var MAX_PATCHLIST_CACHE_SIZE = 20;
var SortedPatchList = /** @class */ (function (_super) {
    __extends(SortedPatchList, _super);
    function SortedPatchList(from_str) {
        var _this = _super.call(this) || this;
        _this.patches = [];
        // the patches indexed by time
        _this.times = {};
        _this.cache = new patch_value_cache_1.PatchValueCache();
        // all the times when there are snapshots.
        _this.all_snapshot_times = {};
        _this.from_str = from_str;
        return _this;
    }
    SortedPatchList.prototype.close = function () {
        this.removeAllListeners();
        misc_1.close(this);
    };
    /* Choose the next available time in ms that is congruent to
       m modulo n.  The congruence condition is so that any time
       collision will have to be with a single person editing a
       document with themselves -- two different users are
       guaranteed to not collide.  Note: even if there is a
       collision, it will automatically fix itself very quickly. */
    SortedPatchList.prototype.next_available_time = function (time, m, n) {
        if (m === void 0) { m = 0; }
        if (n === void 0) { n = 1; }
        var t;
        if (typeof time === "number") {
            t = time;
        }
        else {
            t = time.valueOf();
        }
        if (n <= 0) {
            n = 1;
        }
        var a = m - (t % n);
        if (a < 0) {
            a += n;
        }
        t += a; // now t = m (mod n)
        while (this.times[t] != null) {
            t += n;
        }
        return new Date(t);
    };
    SortedPatchList.prototype.add = function (patches) {
        var e_1, _a;
        if (patches.length === 0) {
            // nothing to do
            return;
        }
        var v = [];
        var oldest = undefined;
        var x;
        var _loop_1 = function () {
            var t = x.time.valueOf();
            var cur = this_1.times[t];
            if (cur != null) {
                // Hmm -- We already have a patch with this time.
                if (underscore_1.isEqual(cur.patch, x.patch) &&
                    cur.user_id === x.user_id &&
                    cur.snapshot === x.snapshot &&
                    misc_1.cmp_Date(cur.prev, x.prev) === 0) {
                    return "continue";
                }
                else {
                    // (1) adding a snapshot or (2) a timestamp collision -- remove duplicate
                    // remove patch with same timestamp from the sorted list of patches
                    this_1.patches = this_1.patches.filter(function (y) { return y.time.valueOf() !== t; });
                    this_1.emit("overwrite", t);
                }
            }
            v.push(x);
            this_1.times[t] = x;
            if (oldest == null || oldest > x.time) {
                oldest = x.time;
            }
            if (x.snapshot != null) {
                this_1.all_snapshot_times[t] = true;
            }
        };
        var this_1 = this;
        try {
            for (var patches_1 = __values(patches), patches_1_1 = patches_1.next(); !patches_1_1.done; patches_1_1 = patches_1.next()) {
                x = patches_1_1.value;
                _loop_1();
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (patches_1_1 && !patches_1_1.done && (_a = patches_1.return)) _a.call(patches_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
        if (oldest != null) {
            // invalidate anything cached back to oldest.
            this.cache.invalidate(oldest);
        }
        // This is O(n*log(n)) where n is the length of this.patches.
        // TODO: Better would probably be an insertion sort, which
        // would be O(m*log(n)) where m=patches.length...
        if (v.length > 0) {
            delete this.versions_cache;
            this.patches = this.patches.concat(v);
            this.patches.sort(util_1.patch_cmp);
        }
    };
    SortedPatchList.prototype.newest_snapshot_time = function () {
        var t0 = 0;
        var t;
        for (t in this.all_snapshot_times) {
            var d = parseInt(t);
            if (d > t0) {
                t0 = d;
            }
        }
        return new Date(t0);
    };
    /*
      value: Return the value of the document at the given (optional)
      point in time.  If the time is given, only include patches up
      to (and including) the given time; otherwise, return the current
      (newest) value.
  
      If force is true, doesn't use snapshot at given input time, even if
      there is one; this is used to update snapshots in case of offline changes
      getting inserted into the changelog.
  
      If without_times is defined, it must be an array of Date objects; in that case
      the current value of the string is computed, but with all the patches
      at the given times in "without_times" ignored.  This is used elsewhere
      as a building block to implement undo.  We do not assume that
      without_times is sorted.
    */
    SortedPatchList.prototype.value = function (time, force, without_times) {
        var e_2, _a, e_3, _b, e_4, _c;
        // oldest time that is skipped:
        var oldest_without_time = undefined;
        // all skipped times.
        var without_times_set = {};
        if (without_times != null) {
            // Process without_times to get a map from time numbers to true.
            if (without_times.length > 0) {
                try {
                    for (var without_times_1 = __values(without_times), without_times_1_1 = without_times_1.next(); !without_times_1_1.done; without_times_1_1 = without_times_1.next()) {
                        var x = without_times_1_1.value;
                        // convert x to number (then string as key)
                        without_times_set[x.valueOf()] = true;
                        if (oldest_without_time == null || x < oldest_without_time) {
                            oldest_without_time = x;
                        }
                    }
                }
                catch (e_2_1) { e_2 = { error: e_2_1 }; }
                finally {
                    try {
                        if (without_times_1_1 && !without_times_1_1.done && (_a = without_times_1.return)) _a.call(without_times_1);
                    }
                    finally { if (e_2) throw e_2.error; }
                }
                if (time != null &&
                    oldest_without_time != null &&
                    time < oldest_without_time) {
                    // requesting value at time before any oldest_without_time, so
                    // is not relevant, so ignore.
                    oldest_without_time = undefined;
                    without_times = undefined;
                    without_times_set = {};
                }
            }
        }
        // we do not discard patch due to prev if prev is before this.
        var prev_cutoff = this.newest_snapshot_time();
        // Determine oldest cached value (undefined if nothing cached)
        var oldest_cached_time = this.cache.oldest_time();
        var value;
        // If the oldest cached value exists and is at least as old as
        // the requested point in time, use it as a base.
        if (oldest_cached_time != null &&
            (time == null || time >= oldest_cached_time) &&
            (oldest_without_time == null || oldest_without_time > oldest_cached_time)) {
            // There is something in the cache, and it is at least as
            // far back in time as the value we want to compute now or
            // any without skips.
            var entry = void 0;
            if (oldest_without_time != null) {
                // true makes "at most" strict, so <.
                entry = this.cache.newest_value_at_most(oldest_without_time, true);
            }
            else {
                entry = this.cache.newest_value_at_most(time);
            }
            if (entry === undefined) {
                throw Error("BUG -- cache should contain a value, but doesn't");
            }
            var start = entry.start, cache_time = entry.time;
            value = entry.value;
            var x = void 0;
            try {
                for (var _d = __values(this.patches.slice(start, this.patches.length)), _e = _d.next(); !_e.done; _e = _d.next()) {
                    x = _e.value;
                    // all patches starting with the cached one
                    if (time != null && x.time > time) {
                        // We are done -- no more patches need to be applied
                        break;
                    }
                    if (x.prev == null ||
                        this.times[x.prev.valueOf()] ||
                        x.prev <= prev_cutoff) {
                        if (oldest_without_time == null ||
                            (oldest_without_time != null &&
                                !without_times_set[x.time.valueOf()])) {
                            // apply patch x to update value to be closer to what we want
                            value = value.apply_patch(x.patch);
                        }
                    }
                    // also record the time of the last patch we applied:
                    cache_time = x.time;
                    start += 1;
                }
            }
            catch (e_3_1) { e_3 = { error: e_3_1 }; }
            finally {
                try {
                    if (_e && !_e.done && (_b = _d.return)) _b.call(_d);
                }
                finally { if (e_3) throw e_3.error; }
            }
            if (oldest_without_time == null &&
                (time == null || start - entry.start >= 10)) {
                // Newest -- or at least 10 patches needed to be applied -- so cache result
                this.cache.include(cache_time, value, start);
                this.cache.prune(MAX_PATCHLIST_CACHE_SIZE); // TODO: include info about size of docs?
            }
        }
        else {
            // Cache is empty or doesn't have anything sufficiently old to be useful.
            // Find the newest snapshot at a time that is <= time.
            value = this.from_str(""); // default in case no snapshots
            var start = 0;
            for (var i = this.patches.length - 1; i >= 0; i--) {
                if ((time == null || this.patches[i].time <= time) &&
                    this.patches[i].snapshot != null) {
                    var x_1 = this.patches[i];
                    if (x_1.snapshot == null) {
                        throw Error("to satisfy typescript");
                    }
                    if (force && misc_1.cmp_Date(x_1.time, time) === 0) {
                        // If force is true we do NOT want to use the existing snapshot, since
                        // the whole point is to force recomputation of it, as it is wrong.
                        // Instead, we'll use the previous snapshot.
                        continue;
                    }
                    // Found a patch with known snapshot that is as old as the time.
                    // This is the base on which we will apply other patches to move forward
                    // to the requested time.
                    value = this.from_str(x_1.snapshot);
                    start = i + 1;
                    break;
                }
            }
            // Apply each of the patches we need to get from
            // value (the last snapshot) to time.
            var cache_time = undefined;
            var cache_start = start;
            var x = void 0;
            try {
                for (var _f = __values(this.patches.slice(start, this.patches.length)), _g = _f.next(); !_g.done; _g = _f.next()) {
                    x = _g.value;
                    if (time != null && x.time > time) {
                        // Done -- no more patches need to be applied
                        break;
                    }
                    // Apply a patch to move us forward.
                    //console.log("applying patch #{i}")
                    if (x.prev == null ||
                        this.times[x.prev.valueOf()] ||
                        x.prev <= prev_cutoff) {
                        if (oldest_without_time == null ||
                            (oldest_without_time != null &&
                                !without_times_set[x.time.valueOf()])) {
                            try {
                                value = value.apply_patch(x.patch);
                            }
                            catch (err) {
                                // See https://github.com/sagemathinc/cocalc/issues/3191
                                // This apply_patch *can* fail in practice due to
                                // a horrible massively nested data structure that appears
                                // due to a bug.  This happened with #3191.  It's better
                                // just skip the patch than to make the project and all
                                // files basically be massively broken!
                                console.warn("WARNING: unable to apply a patch -- skipping it", err);
                            }
                        }
                    }
                    cache_time = x.time;
                    cache_start += 1;
                }
            }
            catch (e_4_1) { e_4 = { error: e_4_1 }; }
            finally {
                try {
                    if (_g && !_g.done && (_c = _f.return)) _c.call(_f);
                }
                finally { if (e_4) throw e_4.error; }
            }
            if (cache_time != null &&
                oldest_without_time == null &&
                (time == null || cache_start - start >= 10)) {
                // Newest -- or at least 10 patches needed to be applied -- so
                // update the cache with our new known value
                this.cache.include(cache_time, value, cache_start);
                this.cache.prune(MAX_PATCHLIST_CACHE_SIZE); // TODO: use info about size of docs?
            }
        }
        // Use the following only for testing/debugging,
        // since it will make everything VERY slow.
        /*
        if (value != null && !value.is_equal(this.value_no_cache(time))) {
            console.warn(`value for time ${time} is wrong!`)
        }
        */
        return value;
    };
    // VERY Slow -- only for consistency checking purposes and debugging.
    // If snapshots=false, don't use snapshots.
    SortedPatchList.prototype.value_no_cache = function (time, snapshots) {
        var e_5, _a;
        if (snapshots === void 0) { snapshots = true; }
        var value = this.from_str(""); // default in case no snapshots
        var start = 0;
        var prev_cutoff = this.newest_snapshot_time();
        if (snapshots && this.patches.length > 0) {
            for (var i = this.patches.length - 1; i >= 0; i--) {
                var x_2 = this.patches[i];
                if ((time == null || x_2.time <= time) && x_2.snapshot != null) {
                    // Found a patch with known snapshot that is as old as the time.
                    // This is the base on which we will apply other patches to move
                    // forward to the requested time.
                    value = this.from_str(x_2.snapshot);
                    start = i + 1;
                    break;
                }
            }
        }
        // Apply each of the patches we need to get from
        // value (the last snapshot) to time.
        var x;
        try {
            for (var _b = __values(this.patches.slice(start, this.patches.length)), _c = _b.next(); !_c.done; _c = _b.next()) {
                x = _c.value;
                if (time != null && x.time > time) {
                    // Done -- no more patches need to be applied
                    break;
                }
                if (x.prev == null ||
                    this.times[x.prev.valueOf()] ||
                    x.prev <= prev_cutoff) {
                    value = value.apply_patch(x.patch);
                }
                else {
                    console.log("skipping patch due to prev", x);
                }
            }
        }
        catch (e_5_1) { e_5 = { error: e_5_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_5) throw e_5.error; }
        }
        return value;
    };
    // For testing/debugging.  Go through the complete patch history and
    // verify that all snapshots are correct (or not -- in which case say so).
    SortedPatchList.prototype.validate_snapshots = function () {
        var e_6, _a;
        if (this.patches.length === 0) {
            return;
        }
        var value;
        var i = 0;
        var x = this.patches[0];
        if (x.snapshot != null) {
            i += 1;
            value = this.from_str(x.snapshot);
        }
        else {
            value = this.from_str("");
        }
        try {
            for (var _b = __values(this.patches.slice(i)), _c = _b.next(); !_c.done; _c = _b.next()) {
                x = _c.value;
                value = value.apply_patch(x.patch);
                if (x.snapshot != null) {
                    var snapshot_value = this.from_str(x.snapshot);
                    if (!value.is_equal(snapshot_value)) {
                        console.log("FAIL (" + x.time + "): at " + i);
                        console.log("diff(snapshot, correct)=");
                        console.log(JSON.stringify(value.make_patch(snapshot_value)));
                    }
                    else {
                        console.log("GOOD (" + x.time + "): snapshot at " + i + " by " + x.user_id);
                    }
                }
                i += 1;
            }
        }
        catch (e_6_1) { e_6 = { error: e_6_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_6) throw e_6.error; }
        }
    };
    // integer index of user who made the patch at given point in time.
    // Throws an exception if there is no patch at that point in time.
    SortedPatchList.prototype.user_id = function (time) {
        var x = this.patch(time);
        if (x == null) {
            throw Error("no patch at " + time);
        }
        return x.user_id;
    };
    // Returns time when patch was sent out, or undefined.  This is
    // ONLY set if the patch was sent at a significantly different
    // time than when it was created, e.g., due to it being offline.
    // Throws an exception if there is no patch at that point in time.
    SortedPatchList.prototype.time_sent = function (time) {
        return this.patch(time).sent;
    };
    // Patch at a given point in time.
    // Throws an exception if there is no patch at that point in time.
    SortedPatchList.prototype.patch = function (time) {
        var p = this.times[time.valueOf()];
        if (p == null) {
            throw Error("no patch at " + time);
        }
        return p;
    };
    SortedPatchList.prototype.versions = function () {
        // Compute and cache result,then return it; result gets cleared when new patches added.
        if (this.versions_cache == null) {
            this.versions_cache = this.patches.map(function (x) { return x.time; });
        }
        return this.versions_cache;
    };
    // Show the history of this document; used mainly for debugging purposes.
    SortedPatchList.prototype.show_history = function (_a) {
        var e_7, _b;
        var _c = _a === void 0 ? {} : _a, milliseconds = _c.milliseconds, trunc = _c.trunc, log = _c.log;
        if (milliseconds === undefined) {
            milliseconds = false;
        }
        if (trunc === undefined) {
            trunc = trunc ? trunc : 80;
        }
        if (log === undefined) {
            log = console.log;
        }
        var i = 0;
        var s;
        var prev_cutoff = this.newest_snapshot_time();
        var x;
        var first_time = true;
        try {
            for (var _d = __values(this.patches), _e = _d.next(); !_e.done; _e = _d.next()) {
                x = _e.value;
                var tm = x.time;
                var tm_show = milliseconds
                    ? tm.valueOf()
                    : tm.toLocaleString();
                log("-----------------------------------------------------\n", i, x.user_id, tm_show, misc_1.trunc_middle(JSON.stringify(x.patch), trunc));
                if (s === undefined) {
                    s = this.from_str(x.snapshot != null ? x.snapshot : "");
                }
                if (first_time && x.snapshot != null) {
                    // do not apply patch no matter what.
                }
                else {
                    if (x.prev == null ||
                        this.times[x.prev.valueOf()] ||
                        x.prev <= prev_cutoff) {
                        s = s.apply_patch(x.patch);
                    }
                    else {
                        log("prev=" + x.prev.valueOf() + " is missing, so not applying this patch");
                    }
                }
                first_time = false;
                log(x.snapshot ? "(SNAPSHOT) " : "           ", misc_1.trunc_middle(s.to_str(), trunc).trim());
                i += 1;
            }
        }
        catch (e_7_1) { e_7 = { error: e_7_1 }; }
        finally {
            try {
                if (_e && !_e.done && (_b = _d.return)) _b.call(_d);
            }
            finally { if (e_7) throw e_7.error; }
        }
    };
    /* This function does not MAKE a snapshot; it just
       returns the time at which we must plan to make a snapshot.
       Returns undefined if do NOT need to make a snapshot soon.
       NOTE: we want to make snapshots as far in the past from
       right now as possible, since users can't insert new offline
       patches *before* the most recent snapshot (in that situation
       all offline work has to get rebased before being inserted
       in history).
  
       RULE 1: If the number of patches since the most recent snapshot
       (or the start of time) is >= 2*interval, we would make a snapshot
       at the patch that is interval steps forward from most recent
       or start of time.
  
       RULE 2: If the sum of the patch sizes since the last
       snapshot (or start of time) exceeds max_size, we make a
       new snapshot at that point in time (starting from the last snapshot)
       when the sum of the patch sizes exceeds max_size.  We do this
       since this is the most canonical choice, in that many distinct
       participants would be mostly likely to make the same choice, which
       increases the chances of avoiding conflicts.  Also, if there are
       tons of big patches, each time there is a new patch that
       gets committed, a new snapshot would get made until no
       more need to be made.  This isn't maximally efficient, in that
       several extra snapshots might get made, but maybe that is OK.
       */
    SortedPatchList.prototype.time_of_unmade_periodic_snapshot = function (interval, max_size) {
        var n = this.patches.length - 1;
        var cur_size = 0;
        for (var i = n; i >= 0; i--) {
            var is_snapshot = this.patches[i].snapshot != null;
            if (!is_snapshot) {
                // add this patch to our size count.  NOTE -- we do not
                // include the snapshot in the size count, since the
                // snapshot already incorporates the patch itself.
                cur_size += this.patches[i].size;
            }
            if (is_snapshot || i == 0) {
                // This is the most recent snapshot or the beginning of time.
                if (i + 2 * interval <= n) {
                    // Time to make a snapshot, based purely on the number
                    // of patches made since the last snapshot (or beginning of time).
                    return this.patches[i + interval].time;
                }
                // No reason to make snapshot based on number of patches.  What about size?
                if (cur_size > max_size) {
                    // Time to make a snapshot, based on the total size since the last
                    // snapshot (or beginning of time).
                    // Make snapshot at first time where max_size exceeded.
                    // We start at i+1 when snapshot below, since the snapshot position
                    // itself includes the patch.
                    var cnt_size = 0;
                    for (var j = is_snapshot ? i + 1 : i; j <= n; j++) {
                        cnt_size += this.patches[j].size;
                        if (cnt_size > max_size) {
                            return this.patches[j].time;
                        }
                    }
                    return; // this should be unreachable
                }
                else {
                    // We found a relatively recent snapshot before max_size exceeded,
                    // so we don't need to make a snapshot.
                    return;
                }
            }
        }
    };
    // Times of all snapshots in memory on this client; these are
    // the only ones we need to worry about for offline patches...
    SortedPatchList.prototype.snapshot_times = function () {
        var v = [];
        var t;
        for (t in this.all_snapshot_times) {
            v.push(new Date(parseInt(t)));
        }
        v.sort(misc_1.cmp_Date);
        return v;
    };
    /* Return the most recent time of a patch, or undefined if
       there are no patches. */
    SortedPatchList.prototype.newest_patch_time = function () {
        if (this.patches.length === 0) {
            return;
        }
        return this.patches[this.patches.length - 1].time;
    };
    SortedPatchList.prototype.count = function () {
        return this.patches.length;
    };
    SortedPatchList.prototype.export = function () {
        return misc_1.deep_copy(this.patches);
    };
    return SortedPatchList;
}(events_1.EventEmitter));
exports.SortedPatchList = SortedPatchList;
