"use strict";
/*
 *  This file is part of CoCalc: Copyright © 2020 Sagemath, Inc.
 *  License: AGPLv3 s.t. "Commons Clause" – see LICENSE.md for details
 */
Object.defineProperty(exports, "__esModule", { value: true });
var patch_value_cache_1 = require("../patch-value-cache");
var doc_1 = require("../../string/doc");
describe("Test out an empty cache", function () {
    var cache;
    it("creates the empty cache", function () {
        cache = new patch_value_cache_1.PatchValueCache();
        expect(typeof cache).toBe("object"); // trivial check
    });
    it("invalidates the empty cache (no op)", function () {
        cache.invalidate(new Date());
    });
    it("prunes the empty cache (no op)", function () {
        cache.prune(0);
    });
    it("gets newest value after some point -- nothing there, so undefined", function () {
        expect(cache.newest_value_at_most()).toBe(undefined);
    });
    it("gets cached entry at given point in time, so undefined", function () {
        expect(cache.get(new Date())).toBe(undefined);
    });
    it("gets oldest cached time, so undefined since there are none", function () {
        expect(cache.oldest_time()).toBe(undefined);
    });
    it("there are no cached values", function () {
        expect(cache.size()).toBe(0);
    });
});
describe("Test a cache with some contents", function () {
    var cache;
    var times = [
        new Date("2019-01-03T20:34"),
        new Date("2019-01-03T20:40"),
        new Date("2019-01-03T22:00"),
    ];
    var values = [
        new doc_1.StringDocument("CoCalc"),
        new doc_1.StringDocument("SageMath"),
        new doc_1.StringDocument("SageMathCloud"),
    ];
    var starts = [3, 10, 20];
    it("creates a cache and populates it", function () {
        cache = new patch_value_cache_1.PatchValueCache();
        for (var n = 0; n <= 2; n++) {
            cache.include(times[n], values[n], starts[n]);
        }
    });
    it("gets oldest cached time", function () {
        expect(cache.oldest_time()).toEqual(times[0]);
    });
    it("there are some cached values", function () {
        expect(cache.size()).toEqual(times.length);
    });
    it("gets newest value after some point", function () {
        for (var n = 0; n <= 2; n++) {
            var v_1 = cache.newest_value_at_most(times[n]);
            expect(v_1 != null).toBe(true);
            if (v_1 != null) {
                expect(v_1.value.to_str()).toBe(values[n].to_str());
            }
        }
        var v = cache.newest_value_at_most(new Date("2019-01-03T20:35"));
        expect(v != null).toBe(true);
        if (v != null) {
            expect(v.value.to_str()).toBe(values[0].to_str());
        }
    });
    it("gets cached entry at given point in time", function () {
        for (var n = 0; n <= 2; n++) {
            var e = cache.get(times[n]);
            expect(e != null).toBe(true);
            if (e != null) {
                expect(e.value.to_str()).toBe(values[n].to_str());
            }
        }
    });
    it("invalidates a cached value", function () {
        // this also invalidates cache for times[2] (which is later),
        // hence the 1 below.
        cache.invalidate(times[1]);
        expect(cache.size()).toBe(1);
    });
    it("prunes the cache down to size 2 (from size 3)...", function () {
        // put the two that gote removed above back:
        for (var n = 1; n <= 2; n++) {
            cache.include(times[n], values[n], starts[n]);
        }
        cache.prune(2);
        expect(cache.size()).toBe(2);
    });
});
