"use strict";
/*
 *  This file is part of CoCalc: Copyright © 2020 Sagemath, Inc.
 *  License: AGPLv3 s.t. "Commons Clause" – see LICENSE.md for details
 */
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
Object.defineProperty(exports, "__esModule", { value: true });
/* Unit test of the SortedPatchList object, which manages
   a sorted list of patches.
*/
var sorted_patch_list_1 = require("../sorted-patch-list");
var util_1 = require("../util");
var doc_1 = require("../../string/doc");
function from_str(s) {
    return new doc_1.StringDocument(s);
}
describe("Test empty sorted patch list -- call all public methods", function () {
    var patches;
    it("creates a sorted patch list", function () {
        patches = new sorted_patch_list_1.SortedPatchList(from_str);
        expect(patches.value().to_str()).toBe("");
    });
    it("calls next_available_time to get next time cong to 3 mod 10", function () {
        var d = new Date(1546543588867); // 7 mod 10.
        var t = patches.next_available_time(d, 3, 10);
        var d1 = t.valueOf();
        expect(d1 % 10).toEqual(3);
        expect(d1 > d.valueOf()).toBe(true);
        expect(d1 - d.valueOf() < 1000).toBe(true);
    });
    it("gets the current value (empty string)", function () {
        expect(patches.value().to_str()).toBe("");
    });
    it("gets the current value without cache", function () {
        expect(patches.value_no_cache().to_str()).toBe("");
    });
    it("validates snapshots (triviality)", function () {
        patches.validate_snapshots();
    });
    it("gets id of user who made edit at time (error since no edits)", function () {
        expect(function () { return patches.user_id(new Date()); }).toThrow("no patch");
    });
    it("gets time sent of a patch (error since no patches)", function () {
        expect(function () { return patches.time_sent(new Date()); }).toThrow("no patch");
    });
    it("gets patch at time (error since no patches)", function () {
        expect(function () { return patches.patch(new Date()); }).toThrow("no patch");
    });
    it("empty list of versions", function () {
        expect(patches.versions()).toEqual([]);
    });
    it("show history doesn't crash", function () {
        patches.show_history({});
    });
    it("time of next snapshot (none since nothing yet)", function () {
        expect(patches.time_of_unmade_periodic_snapshot(100, 9999)).toBe(undefined);
    });
    it("time of snapshots (none of course)", function () {
        expect(patches.snapshot_times()).toEqual([]);
    });
    it("most recent patch", function () {
        expect(patches.newest_patch_time()).toBe(undefined);
    });
    it("number of patch", function () {
        expect(patches.count()).toBe(0);
    });
    it("close it, and check that it is seriously broken", function () {
        patches.close();
        expect(function () { return patches.versions(); }).toThrow("Cannot read property");
    });
});
describe("Test sorted patch list with one patch", function () {
    var patches;
    it("creates a sorted patch list", function () {
        patches = new sorted_patch_list_1.SortedPatchList(from_str);
        expect(patches.value().to_str()).toBe("");
    });
    var patch = {
        time: new Date("2019-01-03T20:33:47.360Z"),
        patch: util_1.make_patch("", "CoCalc"),
        user_id: 0,
        size: JSON.stringify(util_1.make_patch("", "CoCalc")).length,
    };
    it("adds a patch", function () {
        patches.add([patch]);
    });
    it("gets the current value", function () {
        expect(patches.value().to_str()).toBe("CoCalc");
    });
    it("gets the current value again (will use a cache)", function () {
        expect(patches.value().to_str()).toBe("CoCalc");
    });
    it("gets the current value without cache", function () {
        expect(patches.value_no_cache().to_str()).toBe("CoCalc");
    });
    it("gets id of user who made edit at time", function () {
        expect(patches.user_id(patch.time)).toBe(patch.user_id);
    });
    it("gets time sent of a patch (undefined since not yet sent)", function () {
        expect(patches.time_sent(patch.time)).toBe(undefined);
    });
    it("gets patch at time (error since no patches)", function () {
        expect(patches.patch(patch.time).patch).toEqual(patch.patch);
    });
    it("list of versions", function () {
        expect(patches.versions()).toEqual([patch.time]);
    });
    it("most recent patch", function () {
        expect(patches.newest_patch_time()).toEqual(patch.time);
    });
    it("number of patch", function () {
        expect(patches.count()).toBe(1);
    });
});
describe("Test sorted patch list with several patches", function () {
    var patches;
    it("creates a sorted patch list", function () {
        patches = new sorted_patch_list_1.SortedPatchList(from_str);
        expect(patches.value().to_str()).toBe("");
    });
    var w = [
        util_1.make_patch("", "CoCalc"),
        util_1.make_patch("CoCalc", "CoCalc -- Collaborative Calculation"),
        util_1.make_patch("CoCalc -- Collaborative Calculation", "CoCalc -- Collaborative Calculation in the Cloud"),
    ];
    var v = [
        {
            time: new Date("2019-01-03T20:33:47.360Z"),
            patch: w[0],
            user_id: 0,
            sent: new Date("2019-01-03T20:33:47.40Z"),
            size: JSON.stringify(w[0]).length,
        },
        {
            time: new Date("2019-01-03T20:33:50Z"),
            patch: w[1],
            user_id: 1,
            sent: new Date("2019-01-03T20:34"),
            size: JSON.stringify(w[1]).length,
        },
        {
            time: new Date("2019-01-03T20:34:50Z"),
            patch: w[2],
            user_id: 0,
            size: JSON.stringify(w[2]).length,
        },
    ];
    it("adds the patches", function () {
        patches.add(v);
    });
    it("gets the current value", function () {
        expect(patches.value().to_str()).toBe("CoCalc -- Collaborative Calculation in the Cloud");
    });
    it("gets the current value again (will use a cache)", function () {
        expect(patches.value().to_str()).toBe("CoCalc -- Collaborative Calculation in the Cloud");
    });
    it("gets the current value without cache", function () {
        expect(patches.value_no_cache().to_str()).toBe("CoCalc -- Collaborative Calculation in the Cloud");
    });
    it("gets id of user who made edit at time", function () {
        var e_1, _a;
        try {
            for (var v_1 = __values(v), v_1_1 = v_1.next(); !v_1_1.done; v_1_1 = v_1.next()) {
                var patch = v_1_1.value;
                expect(patches.user_id(patch.time)).toBe(patch.user_id);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (v_1_1 && !v_1_1.done && (_a = v_1.return)) _a.call(v_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
    });
    it("gets time sent of patches", function () {
        var e_2, _a;
        try {
            for (var v_2 = __values(v), v_2_1 = v_2.next(); !v_2_1.done; v_2_1 = v_2.next()) {
                var patch = v_2_1.value;
                expect(patches.user_id(patch.time)).toEqual(patch.user_id);
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (v_2_1 && !v_2_1.done && (_a = v_2.return)) _a.call(v_2);
            }
            finally { if (e_2) throw e_2.error; }
        }
    });
    it("gets patch at time", function () {
        var e_3, _a;
        try {
            for (var v_3 = __values(v), v_3_1 = v_3.next(); !v_3_1.done; v_3_1 = v_3.next()) {
                var patch = v_3_1.value;
                expect(patches.patch(patch.time)).toEqual(patch);
            }
        }
        catch (e_3_1) { e_3 = { error: e_3_1 }; }
        finally {
            try {
                if (v_3_1 && !v_3_1.done && (_a = v_3.return)) _a.call(v_3);
            }
            finally { if (e_3) throw e_3.error; }
        }
    });
    it("list of versions", function () {
        expect(patches.versions()).toEqual(v.map(function (x) { return x.time; }));
    });
    it("most recent patch", function () {
        expect(patches.newest_patch_time()).toEqual(v[2].time);
    });
    it("number of patch", function () {
        expect(patches.count()).toBe(v.length);
    });
});
describe("Test inserting missing patches (thus changing history)", function () {
    var patches;
    it("creates a sorted patch list", function () {
        patches = new sorted_patch_list_1.SortedPatchList(from_str);
        expect(patches.value().to_str()).toBe("");
    });
    var w = [
        util_1.make_patch("", "SageMathCloud -- "),
        util_1.make_patch("SageMathCloud -- ", "CoCalc -- "),
        util_1.make_patch("SageMathCloud -- ", "SageMathCloud -- Collaborative Calculation in the Cloud"),
    ];
    var v = [
        {
            time: new Date("2019-01-03T20:33:47.360Z"),
            patch: w[0],
            user_id: 0,
            sent: new Date("2019-01-03T20:33:47.40Z"),
            size: JSON.stringify(w[0]).length,
        },
        {
            time: new Date("2019-01-03T20:33:50Z"),
            patch: w[1],
            user_id: 1,
            sent: new Date("2019-01-03T20:34"),
            size: JSON.stringify(w[1]).length,
        },
        {
            time: new Date("2019-01-03T20:34:50Z"),
            patch: w[2],
            user_id: 0,
            sent: new Date("2019-01-03T20:35"),
            size: JSON.stringify(w[1]).length,
        },
    ];
    it("adds some patches", function () {
        patches.add([v[0], v[2]]);
    });
    it("gets the current value", function () {
        expect(patches.value().to_str()).toBe("SageMathCloud -- Collaborative Calculation in the Cloud");
    });
    it("number of patch", function () {
        expect(patches.count()).toBe(2);
    });
    it("adds other user patch (back in time)", function () {
        patches.add([v[1]]);
    });
    it("number of patch", function () {
        expect(patches.count()).toBe(3);
    });
    it("gets the current value again, which changes", function () {
        expect(patches.value().to_str()).toBe("CoCalc -- Collaborative Calculation in the Cloud");
    });
    it("gets the current value without cache as double check", function () {
        expect(patches.value_no_cache().to_str()).toBe("CoCalc -- Collaborative Calculation in the Cloud");
    });
    it("list of versions", function () {
        expect(patches.versions()).toEqual(v.map(function (x) { return x.time; }));
    });
});
