"use strict";
/*
 *  This file is part of CoCalc: Copyright © 2020 Sagemath, Inc.
 *  License: AGPLv3 s.t. "Commons Clause" – see LICENSE.md for details
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spread = (this && this.__spread) || function () {
    for (var ar = [], i = 0; i < arguments.length; i++) ar = ar.concat(__read(arguments[i]));
    return ar;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.site_license_quota = exports.max_quota = exports.quota = void 0;
/* Compute project quotas based on:

    - settings:
        - by admins
        - system-wide customizable site defaults
        - overcommit ratios
        - maximum overall limits
    - old plans:
        - upgrades coming from about-to-be-deprecated subscriptions and course packages
    - site licenses:
        - old format which is the same as upgrades
        - new format quota object

The quota from coming from settings and user contributions and the
old site license formats all just add together.

The new format SiteLicenseQuota quota object just specifies precisely what the quotas should be
on the project using straightforward human units (with the new quotas all
adding together), rather than adding to some base built in default.  These
have to coexist for about a year.   The way they will combine is that we
first compute the total upgrades coming from everything but the SiteLicenseQuota.
We then compute the total upgrades defined by the SiteLicenseQuota (always
including network automatically, but capping them at whatever maxes we've defined).
Then final quota is the *max* of these two along each parameter.  Old and
new don't add, but of course multiple SiteLicenseQuota do add (since their
costs add).

2021-04-08: a messy "version 1" patchwork code was replaced by a cleaner implementation.
            see rev 19967bb82083b398 for the previous code.
*/
// TODO: relative path just needed in manage-*
var lodash_1 = require("lodash");
var upgrade_spec_1 = require("../upgrade-spec");
var misc_1 = require("../misc");
var MAX_UPGRADES = upgrade_spec_1.upgrades.max_per_project;
// No matter what, every project gets SOME possibly tiny amount of guaranteed cpu.
// This is important since otherwise projects will NOT start at all, e.g., if a paying
// customer is using 100% of the cpu on the node (this will happen if their limits are
// high and they have guaranteed cpu of about 1 or more).  The project will be so slow
// it fails to start in time and times out.
var MIN_POSSIBLE_CPU = Object.freeze({
    member: 0.05,
    nonmember: 0.02,
});
// Min possible **guaranteed** RAM.
var MIN_POSSIBLE_MEMORY = Object.freeze({
    member: 300,
    nonmember: 200,
});
// lower bound for the RAM "limit"
// in particular, we make sure member projects are above the free quota
var MIN_MEMORY_LIMIT = Object.freeze({
    member: 1.5 * upgrade_spec_1.DEFAULT_QUOTAS.memory,
    nonmember: upgrade_spec_1.DEFAULT_QUOTAS.memory,
});
var ZERO_QUOTA = {
    network: false,
    member_host: false,
    privileged: false,
    memory_request: 0,
    cpu_request: 0,
    disk_quota: 0,
    memory_limit: 0,
    cpu_limit: 0,
    idle_timeout: 0,
    always_running: false,
};
// base quota + calculated default quotas is the quota object each project gets by default
// any additional quotas are added on top of it, up until the given limits
var BASE_QUOTAS = {
    network: false,
    member_host: false,
    privileged: false,
    memory_request: 0,
    cpu_request: 0,
    disk_quota: upgrade_spec_1.DEFAULT_QUOTAS.disk_quota,
    memory_limit: upgrade_spec_1.DEFAULT_QUOTAS.memory,
    cpu_limit: upgrade_spec_1.DEFAULT_QUOTAS.cores,
    idle_timeout: upgrade_spec_1.DEFAULT_QUOTAS.mintime,
    always_running: false,
};
// sanitize the overcommitment ratio or discard it
function sanitize_overcommit(oc) {
    if (typeof oc == "number" && !isNaN(oc)) {
        return Math.max(1, oc);
    }
    return undefined;
}
// the quota calculation starts with certain base quotas, which could be modified by the site_settings
function calc_default_quotas(site_settings) {
    var quota = __assign({}, BASE_QUOTAS);
    // overwrite/set extras for any set default quota in the site setting
    if (site_settings != null && site_settings.default_quotas != null) {
        var defaults = site_settings.default_quotas;
        if (typeof defaults.disk_quota == "number") {
            quota.disk_quota = defaults.disk_quota;
        }
        if (typeof defaults.internet == "boolean") {
            quota.network = defaults.internet;
        }
        if (typeof defaults.idle_timeout == "number") {
            quota.idle_timeout = defaults.idle_timeout;
        }
        if (typeof defaults.mem == "number") {
            quota.memory_limit = defaults.mem;
            var oc = sanitize_overcommit(defaults.mem_oc);
            // ratio is 1:mem_oc -- sanitize it first
            if (oc != null) {
                quota.memory_request = Math.round(defaults.mem / oc);
            }
        }
        if (typeof defaults.cpu == "number") {
            quota.cpu_limit = defaults.cpu;
            // ratio is 1:cpu_oc -- sanitize it first
            var oc = sanitize_overcommit(defaults.cpu_oc);
            if (oc != null) {
                quota.cpu_request = defaults.cpu / oc;
            }
        }
    }
    return quota;
}
function isSiteLicenseQuotaSetting(slq) {
    return slq != null && slq.quota != null;
}
function isSettingsQuota(slq) {
    if (slq == null || slq.quota != null)
        return false;
    return (slq.disk_quota != null ||
        slq.memory != null ||
        slq.memory_request != null ||
        slq.cores != null ||
        slq.network != null ||
        slq.cpu_shares != null ||
        slq.mintime != null ||
        slq.member_host != null ||
        slq.ephemeral_state != null ||
        slq.ephemeral_disk != null ||
        slq.always_running != null);
}
// some site licenses do not mix.
// e.g. always_running==true can't upgrade another (especially large) one not having always_running set.
// also preempt upgrades shouldn't uprade member hosting upgades.
//
// this heuristic groups all licenses by always_running and member hosting, and then picks the first nonempty group.
// TODO: once we have a license with an extended uptime (hours instead of ~30 mins), we introduce that as a third dimension.
function select_site_licenses(site_licenses) {
    var e_1, _a;
    var _b, _c;
    if (site_licenses == null || lodash_1.isEmpty(site_licenses))
        return {};
    // key is <member>-<always_running> as 0/1 numbers
    var groups = {
        "0-0": [],
        "0-1": [],
        "1-0": [],
        "1-1": [],
    };
    try {
        // classification
        for (var _d = __values(Object.entries(site_licenses)), _e = _d.next(); !_e.done; _e = _d.next()) {
            var _f = __read(_e.value, 2), key = _f[0], val = _f[1];
            if (val == null)
                continue;
            var is_ar = isSiteLicenseQuotaSetting(val)
                ? val.quota.always_running === true
                : ((_b = val.always_running) !== null && _b !== void 0 ? _b : 0) >= 1;
            var is_member = isSiteLicenseQuotaSetting(val)
                ? val.quota.member === true
                : ((_c = val.member_host) !== null && _c !== void 0 ? _c : 0) >= 1;
            groups[(is_member ? "1" : "0") + "-" + (is_ar ? "1" : "0")].push(key);
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (_e && !_e.done && (_a = _d.return)) _a.call(_d);
        }
        finally { if (e_1) throw e_1.error; }
    }
    // selection -- always_running comes first, then member hosting
    var selected = (function () {
        var e_2, _a, e_3, _b;
        try {
            for (var _c = __values(["1", "0"]), _d = _c.next(); !_d.done; _d = _c.next()) {
                var ar = _d.value;
                try {
                    // always running
                    for (var _e = (e_3 = void 0, __values(["1", "0"])), _f = _e.next(); !_f.done; _f = _e.next()) {
                        var mh = _f.value;
                        // member hosting
                        var k = mh + "-" + ar;
                        if (groups[k].length > 0) {
                            return groups[k];
                        }
                    }
                }
                catch (e_3_1) { e_3 = { error: e_3_1 }; }
                finally {
                    try {
                        if (_f && !_f.done && (_b = _e.return)) _b.call(_e);
                    }
                    finally { if (e_3) throw e_3.error; }
                }
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (_d && !_d.done && (_a = _c.return)) _a.call(_c);
            }
            finally { if (e_2) throw e_2.error; }
        }
    })();
    if (selected == null)
        return {};
    return selected.reduce(function (acc, cur) {
        acc[cur] = site_licenses[cur];
        return acc;
    }, {});
}
// there is an old schema, inherited from SageMathCloud, etc. and newer iterations.
// this helps by going from one schema to the newer one
function upgrade2quota(up) {
    var dflt_false = function (x) {
        return x != null ? (typeof x === "boolean" ? x : to_int(x) >= 1) : false;
    };
    var dflt_num = function (x) {
        return x != null ? (typeof x === "number" ? x : to_float(x)) : 0;
    };
    return {
        network: dflt_false(up.network),
        member_host: dflt_false(up.member_host),
        always_running: dflt_false(up.always_running),
        disk_quota: dflt_num(up.disk_quota),
        memory_limit: dflt_num(up.memory),
        memory_request: dflt_num(up.memory_request),
        cpu_limit: dflt_num(up.cores),
        cpu_request: dflt_num(up.cpu_shares) / 1024,
        privileged: dflt_false(up.privileged),
        idle_timeout: dflt_num(up.mintime),
    };
}
// v2 license upgrades, converting to the quota schema
function license2quota(q) {
    var _a, _b, _c, _d, _e, _f, _g;
    return {
        network: true,
        member_host: !!q.member,
        always_running: !!q.always_running,
        disk_quota: 1000 * ((_a = q.disk) !== null && _a !== void 0 ? _a : 0),
        memory_limit: 1000 * (((_b = q.ram) !== null && _b !== void 0 ? _b : 0) + ((_c = q.dedicated_ram) !== null && _c !== void 0 ? _c : 0)),
        memory_request: 1000 * ((_d = q.dedicated_ram) !== null && _d !== void 0 ? _d : 0),
        cpu_limit: ((_e = q.cpu) !== null && _e !== void 0 ? _e : 0) + ((_f = q.dedicated_cpu) !== null && _f !== void 0 ? _f : 0),
        cpu_request: (_g = q.dedicated_cpu) !== null && _g !== void 0 ? _g : 0,
        privileged: false,
        idle_timeout: 0,
    };
}
// this is summing up several quotas, where we assume they're all fully defined!
function sum_quotas() {
    var e_4, _a;
    var quotas = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        quotas[_i] = arguments[_i];
    }
    var sum = __assign({}, ZERO_QUOTA);
    if (quotas == null || quotas.length == 0)
        return sum;
    try {
        for (var quotas_1 = __values(quotas), quotas_1_1 = quotas_1.next(); !quotas_1_1.done; quotas_1_1 = quotas_1.next()) {
            var q = quotas_1_1.value;
            for (var k in sum) {
                if (typeof sum[k] === "boolean") {
                    sum[k] || (sum[k] = q[k]);
                }
                else {
                    sum[k] += q[k];
                }
            }
        }
    }
    catch (e_4_1) { e_4 = { error: e_4_1 }; }
    finally {
        try {
            if (quotas_1_1 && !quotas_1_1.done && (_a = quotas_1.return)) _a.call(quotas_1);
        }
        finally { if (e_4) throw e_4.error; }
    }
    return sum;
}
// operator for combining two quotas
function op_quotas(q1, q2, op) {
    var e_5, _a;
    var q = {};
    try {
        for (var _b = __values(Object.entries(ZERO_QUOTA)), _c = _b.next(); !_c.done; _c = _b.next()) {
            var _d = __read(_c.value, 2), k = _d[0], v = _d[1];
            if (typeof v === "boolean") {
                q[k] = op === "min" ? q1[k] && q2[k] : q1[k] || q2[k];
            }
            else {
                var cmp = op === "min" ? Math.min : Math.max;
                q[k] = cmp(q1[k], q2[k]);
            }
        }
    }
    catch (e_5_1) { e_5 = { error: e_5_1 }; }
    finally {
        try {
            if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
        }
        finally { if (e_5) throw e_5.error; }
    }
    return q;
}
function min_quotas(q1, q2) {
    return op_quotas(q1, q2, "min");
}
function max_quotas(q1, q2) {
    return op_quotas(q1, q2, "max");
}
// we make sure no matter what, there are certain minimums being set
// still, max upgrades cap those hardcoded minimums
function ensure_minimum(quota, max_upgrades) {
    // ensure minimum cpu are met
    cap_lower_bound(quota, "cpu_request", MIN_POSSIBLE_CPU, max_upgrades);
    // ensure minimum memory request is met
    cap_lower_bound(quota, "memory_request", MIN_POSSIBLE_MEMORY, max_upgrades);
    // ensure minimum memory limit is met
    cap_lower_bound(quota, "memory_limit", MIN_MEMORY_LIMIT, max_upgrades);
    return quota;
}
// if we have some overcommit ratio set, increase a request after we know the quota
// important: the additional requests are capped by an eventually set max_upgrades!
function calc_oc(quota, site_settings, max_upgrades) {
    function ocfun(quota, ratio, limit, request) {
        ratio = sanitize_overcommit(ratio);
        if (ratio != null) {
            var oc_val = quota[limit] / ratio;
            var val = Math.max(quota[request], oc_val);
            if ((max_upgrades === null || max_upgrades === void 0 ? void 0 : max_upgrades[request]) != null) {
                val = Math.min(val, max_upgrades[request]);
            }
            quota[request] = val;
        }
    }
    if ((site_settings === null || site_settings === void 0 ? void 0 : site_settings.default_quotas) != null) {
        var _a = site_settings.default_quotas, mem_oc = _a.mem_oc, cpu_oc = _a.cpu_oc;
        if (quota.cpu_limit != null) {
            ocfun(quota, cpu_oc, "cpu_limit", "cpu_request");
        }
        if (quota.memory_limit != null) {
            ocfun(quota, mem_oc, "memory_limit", "memory_request");
        }
    }
    return quota;
}
// the earlier implementation somehow implicitly rounded down
function round_quota(quota) {
    quota.memory_limit = Math.floor(quota.memory_limit);
    quota.memory_request = Math.floor(quota.memory_request);
    return quota;
}
// calculate how much users can contribute with their upgades
function calc_quota(_a) {
    var e_6, _b;
    var quota = _a.quota, contribs = _a.contribs, max_upgrades = _a.max_upgrades;
    var default_quota = __assign({}, quota);
    // limit the contributions by the overall maximum (except for the defaults!)
    var limited = {};
    try {
        for (var _c = __values(Object.entries(ZERO_QUOTA)), _d = _c.next(); !_d.done; _d = _c.next()) {
            var _e = __read(_d.value, 2), k = _e[0], v = _e[1];
            if (typeof v === "boolean") {
                limited[k] = max_upgrades[k] ? default_quota[k] || contribs[k] : false;
            }
            else {
                var limit = Math.max(0, max_upgrades[k] - default_quota[k]);
                limited[k] = Math.min(contribs[k], limit);
            }
        }
    }
    catch (e_6_1) { e_6 = { error: e_6_1 }; }
    finally {
        try {
            if (_d && !_d.done && (_b = _c.return)) _b.call(_c);
        }
        finally { if (e_6) throw e_6.error; }
    }
    // here we know all fields of the quota are defined
    return limited;
}
function quota_v2(opts) {
    var quota = opts.quota;
    var settings = opts.settings, max_upgrades = opts.max_upgrades, users = opts.users, _a = opts.site_licenses, site_licenses = _a === void 0 ? {} : _a, _b = opts.site_settings, site_settings = _b === void 0 ? {} : _b;
    // limit the default quota by max upgrades
    quota = min_quotas(quota, max_upgrades);
    // classical upgrades by users
    var users_sum = sum_quotas.apply(void 0, __spread(Object.values(users)
        .filter(function (v) { return (v === null || v === void 0 ? void 0 : v.upgrades) != null; })
        .map(function (v) { return upgrade2quota(v.upgrades); })));
    // v1 of licenses, encoding upgrades directly
    var license_upgrades_sum = sum_quotas.apply(void 0, __spread(Object.values(site_licenses).filter(isSettingsQuota).map(upgrade2quota)));
    // v2 of licenses, indirectly via {quota: {…}} objects, introducing yet another schema.
    var license_quota_sum = sum_quotas.apply(void 0, __spread(Object.values(site_licenses)
        .filter(isSiteLicenseQuotaSetting)
        .map(function (l) { return license2quota(l.quota); })));
    // the main idea is old upgrades and licenses quotas only complement each other – not add up.
    // the "settings" object is the "admin upgrade", which isn't capped by the maximum.
    // calc_quota is limited by what's left up until the maximum. also, the previous implementation
    // did add up the deprecated license upgrades as if they were normal user upgrades.
    return ensure_minimum(round_quota(calc_oc(max_quotas(sum_quotas(max_quotas(quota, settings), calc_quota({
        quota: quota,
        contribs: sum_quotas(users_sum, license_upgrades_sum),
        max_upgrades: max_upgrades,
    })), min_quotas(license_quota_sum, max_upgrades)), site_settings, max_upgrades)), max_upgrades);
}
// this is the main function – used by backend services to calculate the run quota of a given project
function quota(settings_arg, users_arg, site_licenses, site_settings) {
    var _a;
    // empirically, this is sometimes a string -- we want this to be a number, though!
    if (typeof (settings_arg === null || settings_arg === void 0 ? void 0 : settings_arg.disk_quota) === "string") {
        settings_arg.disk_quota = to_int(settings_arg.disk_quota);
    }
    // we want to make sure the arguments can't be modified
    var settings = Object.freeze(settings_arg == null ? {} : settings_arg);
    var users = Object.freeze(users_arg == null ? {} : users_arg);
    site_settings = Object.freeze(site_settings);
    // new quota object, we modify it in-place below and return it.
    var quota = calc_default_quotas(site_settings);
    // site settings max quotas overwrite the hardcoded values
    var max_upgrades = Object.freeze(__assign(__assign({}, MAX_UPGRADES), ((_a = site_settings === null || site_settings === void 0 ? void 0 : site_settings.max_upgrades) !== null && _a !== void 0 ? _a : {})));
    // we might not consider all of them!
    site_licenses = Object.freeze(select_site_licenses(site_licenses));
    return quota_v2({
        quota: quota,
        settings: upgrade2quota(settings),
        max_upgrades: upgrade2quota(max_upgrades),
        users: users,
        site_licenses: site_licenses,
        site_settings: site_settings,
    });
}
exports.quota = quota;
// this is used by webapp, not part of this quota calculation, and also not tested
function max_quota(quota, license_quota) {
    var _a, _b;
    for (var field in license_quota) {
        if (license_quota[field] == null)
            continue;
        if (typeof license_quota[field] == "boolean") {
            // boolean
            quota[field] = !!license_quota[field] || !!quota[field];
        }
        else {
            quota[field] = Math.max((_a = license_quota[field]) !== null && _a !== void 0 ? _a : 0, (_b = quota[field]) !== null && _b !== void 0 ? _b : 0);
        }
    }
}
exports.max_quota = max_quota;
// Compute the contribution to quota coming from the quota field of the site licenses.
// This is max'd with the quota computed using settings, the rest of the licenses, etc.
// The given licenses might be a subset of all, because e.g. it's sort of cheating
// to combine memory upgades of member hosting with preempt hosting, or add a small
// always_running license on top of a cheaper but larger member hosting license.
// @see select_site_licenses
//
// this is only used by webapp, not this quota function, and also not tested
function site_license_quota(site_licenses, max_upgrades_param) {
    var e_7, _a;
    // we filter here as well, b/c this function is used elsewhere
    site_licenses = Object.freeze(select_site_licenses(site_licenses));
    // a fallback, should take site settings into account here as well
    var max_upgrades = max_upgrades_param !== null && max_upgrades_param !== void 0 ? max_upgrades_param : MAX_UPGRADES;
    // we start to define a "base" quota, easier to add up everything
    var total_quota = {
        cpu_limit: 0,
        cpu_request: 0,
        memory_limit: 0,
        memory_request: 0,
        disk_quota: 0,
        always_running: false,
        network: false,
        member_host: false,
        privileged: false,
        idle_timeout: 0,
    };
    try {
        for (var _b = __values(Object.values(site_licenses)), _c = _b.next(); !_c.done; _c = _b.next()) {
            var license = _c.value;
            if (!isSiteLicenseQuotaSetting(license))
                continue;
            var quota_1 = license.quota;
            if (quota_1 == null || misc_1.len(quota_1) == 0)
                continue;
            // If there is any nontrivial new quota contribution, then
            // project automatically gets network access... we trust it.
            total_quota.network = true;
            if (quota_1.always_running) {
                total_quota.always_running || (total_quota.always_running = quota_1.always_running);
            }
            if (quota_1.member) {
                total_quota.member_host || (total_quota.member_host = quota_1.member);
            }
            if (quota_1.cpu) {
                total_quota.cpu_limit += quota_1.cpu;
            }
            if (quota_1.ram) {
                total_quota.memory_limit += 1000 * quota_1.ram;
            }
            if (quota_1.dedicated_cpu) {
                total_quota.cpu_request += quota_1.dedicated_cpu;
                // dedicated CPU also contributes to the shared cpu limit:
                total_quota.cpu_limit += quota_1.dedicated_cpu;
            }
            if (quota_1.dedicated_ram) {
                total_quota.memory_request += 1000 * quota_1.dedicated_ram;
                // The dedicated RAM **also** contributes "for free" to the shared RAM
                // which is an upper bound.
                total_quota.memory_limit += 1000 * quota_1.dedicated_ram;
            }
            if (quota_1.disk) {
                total_quota.disk_quota += 1000 * quota_1.disk;
            }
        }
    }
    catch (e_7_1) { e_7 = { error: e_7_1 }; }
    finally {
        try {
            if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
        }
        finally { if (e_7) throw e_7.error; }
    }
    var ret = limit_quota(total_quota, max_upgrades);
    //console.log("total_quota_limited", JSON.stringify(ret, null, 2));
    return ret;
}
exports.site_license_quota = site_license_quota;
/*
for better understanding of the next two functions, here are two examples

total_quota =  {            max_upgrades = {
  cpu_limit: 3,               disk_quota: 20000,
  cpu_request: 1,             memory: 16000,
  memory_limit: 3000,         memory_request: 8000,
  memory_request: 1000,       cores: 3,
  disk_quota: 2000,           network: 1,
  always_running: true,       cpu_shares: 2048,
  network: true,              mintime: 7776000,
  member_host: true,          member_host: 1,
  privileged: false,          ephemeral_state: 1,
  idle_timeout: 0             ephemeral_disk: 1,
}                             always_running: 1
                            }
*/
function limit_quota(total_quota, max_upgrades) {
    var e_8, _a;
    try {
        for (var _b = __values(Object.entries(upgrade2quota(max_upgrades))), _c = _b.next(); !_c.done; _c = _b.next()) {
            var _d = __read(_c.value, 2), key = _d[0], val = _d[1];
            if (typeof val === "boolean") {
                total_quota[key] && (total_quota[key] = val);
            }
            else {
                total_quota[key] = Math.min(total_quota[key], val);
            }
        }
    }
    catch (e_8_1) { e_8 = { error: e_8_1 }; }
    finally {
        try {
            if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
        }
        finally { if (e_8) throw e_8.error; }
    }
    return total_quota;
}
// TODO name is <K extends keyof Quota>, but that causes troubles ...
// at this point we already know that we only look for numeric properties and they're all != null
function cap_lower_bound(quota, name, MIN_SPEC, max_upgrades) {
    var val = quota[name];
    if (typeof val === "number") {
        var cap = quota.member_host ? MIN_SPEC.member : MIN_SPEC.nonmember;
        if (max_upgrades != null) {
            var max = max_upgrades[name];
            if (typeof max === "number")
                cap = Math.min(cap, max);
        }
        // @ts-ignore
        quota[name] = Math.max(val, cap);
    }
}
function make_number_parser(fn) {
    return function (s) {
        if (s == null)
            return 0;
        try {
            var n = fn(s);
            if (isNaN(n)) {
                return 0;
            }
            else {
                return n;
            }
        }
        catch (error) {
            return 0;
        }
    };
}
var to_int = make_number_parser(parseInt);
var to_float = make_number_parser(parseFloat);
